#include "StdAfx.h"
#if WAR_RFC2068
#include "WarSvrProtocolHttp.h"   // class implemented

#ifndef WAR_LOG_H
#   include "WarLog.h"
#endif

#ifndef WAR_AUTO_LOCK_H
#   include "WarAutoLock.h"
#endif

#ifndef WAR_SVR_ENGINE_H
#   include "WarSvrEngine.h"
#endif

#ifndef WAR_TIMER_H
#   include "WarTimer.h"
#endif

#ifndef WAR_PERFMON_DEF_H
#   include "WarPerfmonDef.h"
#endif

#ifndef WAR_USER_AUTH_IMPERSONATE_H
#   include "WarUserAuthImpersonate.h"
#endif

#ifndef WAR_ATOI_H
#   include "WarAtoi.h"
#endif

#define AUTO_LOCK WarAutoLock MyLock(mSvrLock);

using namespace std;

/////////////////////////////// PUBLIC ///////////////////////////////////////

//============================= LIFECYCLE ====================================

WarSvrProtocolHttp::WarSvrProtocolHttp(
        war_socket_io_ptr_t& companionPtr,
        WarSvrDefinition& svrDefinion) :
WarSvrProtocol(companionPtr, svrDefinion)
{
    mReply.mpProtocol = this;
    Reset();
}

//============================= OPERATORS ====================================
//============================= OPERATIONS ===================================

void WarSvrProtocolHttp::Abort(const WarError& reason)
        throw(WarException)
{
    CloseConnection(reason);
}

void WarSvrProtocolHttp::CloseConnection(const WarError& reason) 
        throw(WarException)
{
    if (IsOpen())
    {
        Close();
    }

    Logout();
    GetEngine().UnregisterSession(*this);
    WarThrow(reason, NULL);
}

void WarSvrProtocolHttp::Reset()
{
    mReply.Reset();
    mHaveEndOfHeader = false;
    if (!mInputBuffer.IsEmpty())
        mInputBuffer.Reset();

    mExpectedContentBytes = 0;
    mReceivedContentBytes = 0;
}

//============================= CALLBACK   ===================================

void WarSvrProtocolHttp::OnClientConnect()
{
    AUTO_LOCK
    // Update internal variables
    // and start receiving commands from the user

    try
    {
        WarSvrProtocol::OnClientConnect();
    }
    catch(WarException& ex)
    {
        switch(ex.LocalError())
        {
        case WAR_ERR_IP_IS_DENIED:
            mReply << HSM_BAD_IP << HTTPR_FORBIDDEN;
            break;
        case WAR_ERR_ACCESS_DENIED:
            mReply << HSM_ACCESS_DENIED << HTTPR_FORBIDDEN;
            break;
        default:
            mReply << HSM_GENERIC_FAIL << HTTPR_INTERNAL_SERVER_ERROR;
            break;
        }

        CloseConnection(ex); // Throws 
    }

    WarUrl my_url;
    my_url.Create("http://", NULL, NULL, 
        GetLocalAddress().GetPort(), NULL, "/");

    mRequestHeader.Reset();
    mRequestHeader.SetDefaultUrl(my_url);
}

void WarSvrProtocolHttp::OnReceived(const WarError& status, 
        war_transfer_buffer_ptr_t& buffer)
{
    try
    {
        ProcessInput(status, buffer);
    }
    catch(WarException& ex)
    {
        // Exception during input. This will normally indicate 
        // an error-situation where we should give a short
        // explanation and close the connection.
        // Exceptions are when we negotiate authentication
        // credentials.
        switch(ex.LocalError())
        {
        case WAR_HTERR_HEADER_TOO_LONG:
            GetReply() << HSM_HEADER_TOO_LONG 
                << HTTPR_INTERNAL_SERVER_ERROR;
            Close();
            return;
        case WAR_ERR_NOT_IMPLEMENTED:
            GetReply() << HSM_NOT_IMPLEMENTED 
                << HTTPR_NOT_IMPLEMENTED;
            Close();
            return;
        }
    }
}

void WarSvrProtocolHttp::OnSent(const WarError& status, 
        war_transfer_buffer_ptr_t& buffer)
{
}

void WarSvrProtocolHttp::OnValidateHeaderLength(size_t currLen)
throw(WarException)
{
    size_t max_len;
    if (currLen > (max_len = (size_t)GetIntOption("http_MAXHDRLEN")))
    {
        WarLog security_log(WARLOG_SECURITY, "WarSvrProtocolHttp::OnValidateHeaderLength()");

        security_log << "HTTP header length exeeds the limit of "
            << max_len
            << "bytes defined by http_MAXHDRLEN option. "
            "This may indicate a breakin attempt from a hostile host."
            << war_endl;

        WarThrow(WarError(WAR_HTERR_HEADER_TOO_LONG), NULL);
    }
}

//============================= ACCESS     ===================================
//============================= INQUIRY    ===================================
/////////////////////////////// PROTECTED  ///////////////////////////////////

WarSvrProtocolHttp::~WarSvrProtocolHttp(void)
{
}

void WarSvrProtocolHttp::ProcessInput(const WarError& status, 
        war_transfer_buffer_ptr_t& buffer)
{
    mInputBuffer.PushData(buffer->GetBuffer(), 
        buffer->GetBytesUsed());

again:

    if (!mHaveEndOfHeader)
    {
        // Scan for end of header marker
        war_ccstr_t p = mInputBuffer.GetStartOfBuffer();
        war_ccstr_t pp = mInputBuffer.GetEndOfBuffer() -3;

        while(p < pp)
        {
            if ((p[0] == '\r')
                && (p[1] == '\n')
                && (p[2] == '\r')
                && (p[3] == '\n'))
            {
                mHaveEndOfHeader = true;
                size_t header_len = (p - mInputBuffer.GetStartOfBuffer()) + 4;

                // Check if the header-length is too large
                OnValidateHeaderLength(header_len);

                // Process header
                mRequestHeader << mInputBuffer.PopData(header_len);
                mReply.GetHeader().SetHttpVersion(mRequestHeader.GetHttpVersion());

                // Load properties for the virtual host, if present
                LoadVirtual(mRequestHeader.GetUrl());

                // Verify IP Access
                VerifyRemoteIpAddress();

                // Attempt to log on
                VerifyHttpAuth();

                // Check if method is allowed
                VerifyHttpRequest();

                // Process headers
                try
                {
                    mExpectedContentBytes = WarAtoi<contentlen_t>(
                        mRequestHeader.GetHeaderValue("Content-Length").c_str());
                }
                catch(WarException)
                {
                }

                goto process_content;
            }
            else
                ++p;
        }

        // Check if the header-length is too large
        OnValidateHeaderLength(mInputBuffer.GetLength());

        if (!mHaveEndOfHeader)
            return; // Need more input
    }

process_content:
    
    mReceivedContentBytes += mInputBuffer.GetLength();
    
    if (mExpectedContentBytes >= mReceivedContentBytes)
    {
        // The request is complete
        war_ccstr_t data = mInputBuffer.PopData(mExpectedContentBytes);
        WarError caught_err;

        try
        {
            switch(mRequestHeader.GetMethod())
            {
            case HM_GET:
                OnGetRequest(data, mExpectedContentBytes);
                break;
            case HM_OPTIONS:
                OnOptionsRequest(data, mExpectedContentBytes);
                break;
            case HM_HEAD:
                OnHeadRequest(data, mExpectedContentBytes);
                break;
            case HM_POST:
                OnPostRequest(data, mExpectedContentBytes);
                break;
            case HM_PUT:
                OnPutRequest(data, mExpectedContentBytes);
                break;
            case HM_DELETE:
                OnDeleteRequest(data, mExpectedContentBytes);
                break;
            case HM_TRACE:
                OnTraceRequest(data, mExpectedContentBytes);
                break;
            default:
                OnInvalidRequest(data, mExpectedContentBytes);
            }
        }
        catch(WarException &ex)
        {
            if (ex.LocalError() == WAR_ERR_NOT_IMPLEMENTED)
            {
                WarLog db_log(WARLOG_DEBUG, "WarSvrProtocolHttp::ProcessInput()");
                if (db_log)
                {
                    db_log << "HTTP request type "
                        << mRequestHeader.GetMethodName()
                        << " not supported (uninplemented)"
                        << war_endl;
                }

                throw ex;
            }

            WarLog warn_log(WARLOG_WARNINGS, "WarSvrProtocolHttp::ProcessInput()");

            warn_log << "Caught an unexcpected exception while "
                "processing a HTTP request. "
                << ex
                << war_endl;
        }

        // Prepare for the next request

        if (mReply.GetHeader().GetHttpVersion() == HTTP_1_0)
        {
            Close();
            return; // HTTP 1.0 can only do one request/connection.
        }

        Reset();

        if (!buffer.IsEmpty())
            goto again; // Asynch requests
    }
}


void WarSvrProtocolHttp::OnGetRequest(war_ccstr_t contentData, 
                                      const contentlen_t contentLength)
                                      throw(WarException)
{
    WarThrow(WarError(WAR_ERR_NOT_IMPLEMENTED), NULL);
}

void WarSvrProtocolHttp::OnOptionsRequest(war_ccstr_t contentData, 
                                          const contentlen_t contentLength)
                                          throw(WarException)
{
    WarThrow(WarError(WAR_ERR_NOT_IMPLEMENTED), NULL);
}

void WarSvrProtocolHttp::OnHeadRequest(war_ccstr_t contentData, 
                                       const contentlen_t contentLength)
                                       throw(WarException)
{
    WarThrow(WarError(WAR_ERR_NOT_IMPLEMENTED), NULL);
}

void WarSvrProtocolHttp::OnPostRequest(war_ccstr_t contentData, 
                                       const contentlen_t contentLength)
                                       throw(WarException)
{
    WarThrow(WarError(WAR_ERR_NOT_IMPLEMENTED), NULL);
}

void WarSvrProtocolHttp::OnPutRequest(war_ccstr_t contentData, 
                                      const contentlen_t contentLength)
                                      throw(WarException)
{
    WarThrow(WarError(WAR_ERR_NOT_IMPLEMENTED), NULL);
}

void WarSvrProtocolHttp::OnDeleteRequest(war_ccstr_t contentData, 
                                         const contentlen_t contentLength)
                                         throw(WarException)
{
    WarThrow(WarError(WAR_ERR_NOT_IMPLEMENTED), NULL);
}

void WarSvrProtocolHttp::OnTraceRequest(war_ccstr_t contentData, 
                                        const contentlen_t contentLength)
                                        throw(WarException)
{
    WarThrow(WarError(WAR_ERR_NOT_IMPLEMENTED), NULL);
}

void WarSvrProtocolHttp::OnInvalidRequest(war_ccstr_t contentData, 
                                          const contentlen_t contentLength)
                                          throw(WarException)
{
    WarThrow(WarError(WAR_ERR_NOT_IMPLEMENTED), NULL);
}

/////////////////////////////// PRIVATE    ///////////////////////////////////

#endif // #if WAR_RFC2068

